# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API client."""
from __future__ import unicode_literals

import os
import logging
import sys

# pylint: disable=wrong-import-order
import bs4
import requests

# pylint: disable=redefined-builtin
from requests.exceptions import ConnectionError
from urllib3.exceptions import InsecureRequestWarning
import webbrowser

# pylint: disable-msg=import-error
from google_auth_oauthlib import flow as googleauth_flow
import google.auth.transport.requests
import pandas

from . import credentials
from . import definitions
from . import error
from . import index
from . import sketch
from . import user
from . import version
from . import sigma


logger = logging.getLogger("timesketch_api.client")


class TimesketchApi:
    """Timesketch API object

    Attributes:
        api_root: The full URL to the server API endpoint.
        session: Authenticated HTTP session.
    """

    DEFAULT_OAUTH_SCOPE = [
        "https://www.googleapis.com/auth/userinfo.email",
        "openid",
        "https://www.googleapis.com/auth/userinfo.profile",
    ]

    DEFAULT_OAUTH_AUTH_URL = "https://accounts.google.com/o/oauth2/v2/auth"
    DEFAULT_OAUTH_TOKEN_URL = "https://oauth2.googleapis.com/token"
    DEFAULT_OAUTH_PROVIDER_URL = "https://www.googleapis.com/oauth2/v1/certs"
    DEFAULT_OAUTH_LOCALHOST_URL = "http://localhost"
    DEFAULT_OAUTH_API_CALLBACK = "/login/api_callback/"

    # Default retry count for operations that attempt a retry.
    DEFAULT_RETRY_COUNT = 5

    def __init__(
        self,
        host_uri,
        username,
        password="",
        verify=True,
        client_id="",
        client_secret="",
        auth_mode="userpass",
        create_session=True,
    ):
        """Initializes the TimesketchApi object.

        Args:
            host_uri: URI to the Timesketch server (https://<server>/).
            username: User username.
            password: User password.
            verify: Verify server SSL certificate.
            client_id: The client ID if OAUTH auth is used.
            client_secret: The OAUTH client secret if OAUTH is used.
            auth_mode: The authentication mode to use. Defaults to 'userpass'
                Supported values are 'userpass' (username/password combo),
                'http-basic' (HTTP Basic authentication) and oauth.
            create_session: Boolean indicating whether the client object
                should create a session object. If set to False the
                function "set_session" needs to be called before proceeding.

        Raises:
            ConnectionError: If the Timesketch server is unreachable.
            RuntimeError: If the client is unable to authenticate to the
                backend.
        """
        self._host_uri = host_uri
        self.api_root = "{0:s}/api/v1".format(host_uri)
        self.credentials = None
        self._flow = None

        if not create_session:
            self._session = None
            return

        try:
            self._session = self._create_session(
                username,
                password,
                verify=verify,
                client_id=client_id,
                client_secret=client_secret,
                auth_mode=auth_mode,
            )
        except ConnectionError as exc:
            raise ConnectionError("Timesketch server unreachable") from exc
        except RuntimeError as e:
            raise RuntimeError(
                "Unable to connect to server, error: {0!s}".format(e)
            ) from e

    @property
    def current_user(self):
        """Property that returns the user object of the logged in user."""
        return user.User(self)

    @property
    def version(self):
        """Property that returns back the API client version."""
        version_dict = self.fetch_resource_data("version/")
        ts_version = None
        if version_dict:
            ts_version = version_dict.get("meta", {}).get("version")

        if ts_version:
            return "API Client: {0:s}\nTS Backend: {1:s}".format(
                version.get_version(), ts_version
            )

        return "API Client: {0:s}".format(version.get_version())

    @property
    def session(self):
        """Property that returns the session object."""
        if self._session is None:
            raise ValueError("Session is not set.")
        return self._session

    def set_credentials(self, credential_object):
        """Sets the credential object."""
        self.credentials = credential_object

    def set_session(self, session_object):
        """Sets the session object."""
        self._session = session_object

    def _authenticate_session(self, session, username, password):
        """Post username/password to authenticate the HTTP session.

        Args:
            session: Instance of requests.Session.
            username: User username.
            password: User password.
        """
        # Do a POST to the login handler to set up the session cookies
        data = {"username": username, "password": password}
        session.post("{0:s}/login/".format(self._host_uri), data=data)

    def _set_csrf_token(self, session):
        """Retrieve CSRF token from the server and append to HTTP headers.

        Args:
            session: Instance of requests.Session.
        """
        # Scrape the CSRF token from the response
        response = session.get(self._host_uri)
        soup = bs4.BeautifulSoup(response.text, features="html.parser")

        tag = soup.find(id="csrf_token")
        csrf_token = None
        if tag:
            csrf_token = tag.get("value")
        else:
            tag = soup.find("meta", attrs={"name": "csrf-token"})
            if tag:
                csrf_token = tag.attrs.get("content")

        if not csrf_token:
            return

        session.headers.update({"x-csrftoken": csrf_token, "referer": self._host_uri})

    def _create_oauth_session(
        self,
        client_id="",
        client_secret="",
        client_secrets_file=None,
        host="localhost",
        port=8080,
        open_browser=False,
        run_server=True,
        skip_open=False,
    ):
        """Return an OAuth session.

        Args:
            client_id: The client ID if OAUTH auth is used.
            client_secret: The OAUTH client secret if OAUTH is used.
            client_secrets_file: Path to the JSON file that contains the client
                secrets, in the client_secrets format.
            host: Host address the OAUTH web server will bind to.
            port: Port the OAUTH web server will bind to.
            open_browser: A boolean, if set to false (default) a browser window
                will not be automatically opened.
            run_server: A boolean, if set to true (default) a web server is
                run to catch the OAUTH request and response.
            skip_open: A booelan, if set to True (defaults to False) an
                authorization URL is printed on the screen to visit. This is
                only valid if run_server is set to False.

        Return:
            session: Instance of requests.Session.

        Raises:
            RuntimeError: if unable to log in to the application.
        """
        if client_secrets_file:
            if not os.path.isfile(client_secrets_file):
                raise RuntimeError(
                    "Unable to log in, client secret files does not exist."
                )
            flow = googleauth_flow.InstalledAppFlow.from_client_secrets_file(
                client_secrets_file,
                scopes=self.DEFAULT_OAUTH_SCOPE,
                autogenerate_code_verifier=True,
            )
        else:
            provider_url = self.DEFAULT_OAUTH_PROVIDER_URL
            client_config = {
                "installed": {
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "auth_uri": self.DEFAULT_OAUTH_AUTH_URL,
                    "token_uri": self.DEFAULT_OAUTH_TOKEN_URL,
                    "auth_provider_x509_cert_url": provider_url,
                    "redirect_uris": [self.DEFAULT_OAUTH_LOCALHOST_URL],
                },
            }

            flow = googleauth_flow.InstalledAppFlow.from_client_config(
                client_config,
                self.DEFAULT_OAUTH_SCOPE,
                autogenerate_code_verifier=True,
            )

            flow.redirect_uri = self.DEFAULT_OAUTH_LOCALHOST_URL

        if run_server:
            _ = flow.run_local_server(host=host, port=port, open_browser=open_browser)
        else:
            if not sys.stdout.isatty() or not sys.stdin.isatty():
                msg = (
                    "You will be asked to paste a token into this session to"
                    "authenticate, but the session doesn't have a tty"
                )
                raise RuntimeError(msg)

            auth_url, _ = flow.authorization_url(prompt="select_account")

            if skip_open:
                print("Visit the following URL to authenticate: {0:s}".format(auth_url))
            else:
                open_browser = input("Open the URL in a browser window? [y/N] ")
                if open_browser.lower() == "y" or open_browser.lower() == "yes":
                    webbrowser.open(auth_url)
                else:
                    print(
                        "Need to manually visit URL to authenticate: "
                        "{0:s}".format(auth_url)
                    )

            code = input("Enter the token code: ")
            _ = flow.fetch_token(code=code)

        session = flow.authorized_session()
        self._flow = flow
        self.credentials = credentials.TimesketchOAuthCredentials()
        self.credentials.credential = flow.credentials
        return self.authenticate_oauth_session(session)

    def authenticate_oauth_session(self, session):
        """Authenticate an OAUTH session.

        Args:
            session: Authorized session object.
        """
        # Authenticate to the Timesketch backend.
        login_callback_url = "{0:s}{1:s}".format(
            self._host_uri, self.DEFAULT_OAUTH_API_CALLBACK
        )
        params = {
            "id_token": session.credentials.id_token,
        }
        response = session.get(login_callback_url, params=params)
        if response.status_code not in definitions.HTTP_STATUS_CODE_20X:
            error.error_message(
                response, message="Unable to authenticate", error=RuntimeError
            )

        self._set_csrf_token(session)
        return session

    def _create_session(
        self, username, password, verify, client_id, client_secret, auth_mode
    ):
        """Create authenticated HTTP session for server communication.

        Args:
            username: User to authenticate as.
            password: User password.
            verify: Verify server SSL certificate.
            client_id: The client ID if OAUTH auth is used.
            client_secret: The OAUTH client secret if OAUTH is used.
            auth_mode: The authentication mode to use. Supported values are
                'userpass' (username/password combo), 'http-basic'
                (HTTP Basic authentication) and oauth.

        Returns:
            Instance of requests.Session.
        """
        if auth_mode == "oauth":
            return self._create_oauth_session(client_id, client_secret)

        if auth_mode == "oauth_local":
            return self._create_oauth_session(
                client_id=client_id,
                client_secret=client_secret,
                run_server=False,
                skip_open=True,
            )

        session = requests.Session()

        # If using HTTP Basic auth, add the user/pass to the session
        if auth_mode == "http-basic":
            session.auth = (username, password)

        # SSL Cert verification is turned on by default.
        if not verify:
            session.verify = False
            # disable warnings, since user actively decided to set verify to false
            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        # Get and set CSRF token and authenticate the session if appropriate.
        self._set_csrf_token(session)
        if auth_mode == "userpass":
            self._authenticate_session(session, username, password)

        return session

    def fetch_resource_data(self, resource_uri, params=None):
        """Make a HTTP GET request.

        Args:
            resource_uri: The URI to the resource to be fetched.
            params: Dict of URL parameters to send in the GET request.

        Returns:
            Dictionary with the response data.

        Raises:
            ValueError: If response could not be JSON-decoded after
                DEFAULT_RETRY_COUNT attempts.
            RuntimeError: If the API server returns an error or empty.
        """
        resource_url = "{0:s}/{1:s}".format(self.api_root, resource_uri)

        retry_count = 0
        result = None
        while True:
            retry_count += 1
            response = self.session.get(resource_url, params=params)
            try:
                result = error.get_response_json(response, logger)
                if result:
                    return result
            except RuntimeError as e:
                if retry_count >= self.DEFAULT_RETRY_COUNT:
                    raise RuntimeError(
                        "Error for request '{0:s}' - '{1!s}'".format(resource_url, e)
                    ) from e

                logger.warning(
                    "[{0:d}/{1:d}] Parsing the repsonse for request '{2:s}'"
                    "failed. Trying again...".format(
                        retry_count, self.DEFAULT_RETRY_COUNT, resource_url
                    )
                )
            except ValueError as e:
                if retry_count >= self.DEFAULT_RETRY_COUNT:
                    raise ValueError(
                        "Error parsing response for request '{0:s}' - {1!s}".format(
                            resource_url, e
                        )
                    ) from e

                logger.warning(
                    "[{0:d}/{1:d}] Parsing the JSON repsonse for request "
                    "'{2:s}' failed. Trying again...".format(
                        retry_count, self.DEFAULT_RETRY_COUNT, resource_url
                    )
                )

            if retry_count >= self.DEFAULT_RETRY_COUNT:
                raise RuntimeError(
                    "Unable to fetch JSON resource data for request: '{0:s}'"
                    " - Response: '{1!s}'".format(resource_url, result)
                )

    def create_sketch(self, name, description=None):
        """Create a new sketch.

        Args:
            name: Name of the sketch.
            description: Description of the sketch.

        Returns:
            Instance of a Sketch object.

        Raises:
            RuntimeError: If response does not contain an 'objects' key after
                DEFAULT_RETRY_COUNT attempts.
        """
        if not description:
            description = name

        retry_count = 0
        objects = None
        while True:
            resource_url = "{0:s}/sketches/".format(self.api_root)
            form_data = {"name": name, "description": description}
            response = self.session.post(resource_url, json=form_data)
            response_dict = error.get_response_json(response, logger)
            objects = response_dict.get("objects")
            if objects:
                break
            retry_count += 1

            if retry_count >= self.DEFAULT_RETRY_COUNT:
                raise RuntimeError("Unable to create a new sketch.")

        sketch_id = objects[0]["id"]
        return self.get_sketch(sketch_id)

    def create_user(self, username, password):
        """Create a new user.

        Args:
            username: Name of the user
            password: Password of the user

        Returns:
            True if user created successfully.

        Raises:
            RuntimeError: If response does not contain an 'objects' key after
                DEFAULT_RETRY_COUNT attempts.
        """

        retry_count = 0
        objects = None
        while True:
            resource_url = "{0:s}/users/".format(self.api_root)
            form_data = {"username": username, "password": password}
            response = self.session.post(resource_url, json=form_data)
            response_dict = error.get_response_json(response, logger)
            objects = response_dict.get("objects")
            if objects:
                break
            retry_count += 1

            if retry_count >= self.DEFAULT_RETRY_COUNT:
                raise RuntimeError("Unable to create a new user.")

        return user.User(user_id=objects[0]["id"], api=self)

    def list_users(self):
        """Get a list of all users.

        Yields:
            User object instances.
        """
        response = self.fetch_resource_data("users/")

        for user_dict in response.get("objects", [])[0]:
            user_id = user_dict["id"]
            user_obj = user.User(user_id=user_id, api=self)
            yield user_obj

    def get_user(self, user_id):
        """Get a user.

        Args:
            user_id: Primary key ID of the user.

        Returns:
            Instance of a User object.
        """
        return user.User(user_id=user_id, api=self)

    def get_oauth_token_status(self):
        """Return a dict with OAuth token status, if one exists."""
        if not self.credentials:
            return {"status": "No stored credentials."}
        return {
            "expired": self.credentials.credential.expired,
            "expiry_time": self.credentials.credential.expiry.isoformat(),
        }

    def get_sketch(self, sketch_id):
        """Get a sketch.

        Args:
            sketch_id: Primary key ID of the sketch.

        Returns:
            Instance of a Sketch object.
        """
        return sketch.Sketch(sketch_id, api=self)

    def get_aggregator_info(self, name="", as_pandas=False):
        """Returns information about available aggregators.

        Args:
            name: String with the name of an aggregator. If the name is not
                provided, a list with all aggregators is returned.
            as_pandas: Boolean indicating that the results will be returned
                as a Pandas DataFrame instead of a list of dicts.

        Returns:
            A list with dict objects with the information about aggregators,
            unless as_pandas is set, then the function returns a DataFrame
            object.
        """
        resource_uri = "aggregation/info/"

        if name:
            data = {"aggregator": name}
            resource_url = "{0:s}/{1:s}".format(self.api_root, resource_uri)
            response = self.session.post(resource_url, json=data)
            response_json = error.get_response_json(response, logger)
        else:
            response_json = self.fetch_resource_data(resource_uri)

        if not as_pandas:
            return response_json

        lines = []
        if isinstance(response_json, dict):
            response_json = [response_json]

        for line in response_json:
            line_dict = {
                "name": line.get("name", "N/A"),
                "description": line.get("description", "N/A"),
            }
            for field_index, field in enumerate(line.get("fields", [])):
                line_dict["field_{0:d}_name".format(field_index + 1)] = field.get(
                    "name"
                )
                line_dict["field_{0:d}_description".format(field_index + 1)] = (
                    field.get("description")
                )
            lines.append(line_dict)

        return pandas.DataFrame(lines)

    def list_sketches(self, per_page=50, scope="user", include_archived=True):
        """Get a list of all open sketches that the user has access to.

        Args:
            per_page: Number of items per page when paginating. Default is 50.
            scope: What scope to get sketches as. Default to user.
                user: sketches owned by the user
                recent: sketches that the user has actively searched in
                shared: Get sketches that can be accessed
                admin: Get all sketches if the user is an admin
                archived: get archived sketches
                search: pass additional search query
            include_archived: If archived sketches should be returned.

        Yields:
            Sketch objects instances.
        """
        url_params = {
            "per_page": per_page,
            "scope": scope,
            "include_archived": include_archived,
        }
        # Start with the first page
        page = 1
        has_next_page = True

        while has_next_page:
            url_params["page"] = page
            response = self.fetch_resource_data("sketches/", params=url_params)
            meta = response.get("meta", {})

            page = meta.get("next_page")
            if not page:
                has_next_page = False

            for sketch_dict in response.get("objects", []):
                sketch_id = sketch_dict["id"]
                sketch_name = sketch_dict["name"]
                sketch_obj = sketch.Sketch(
                    sketch_id=sketch_id, api=self, sketch_name=sketch_name
                )
                yield sketch_obj

    def get_searchindex(self, searchindex_id):
        """Get a searchindex.

        Args:
            searchindex_id: Primary key ID of the searchindex.

        Returns:
            Instance of a SearchIndex object.
        """
        return index.SearchIndex(searchindex_id, api=self)

    def check_celery_status(self, job_id=""):
        """Return information about outstanding celery tasks or a specific one.

        Args:
            job_id (str): Optional Celery job identification string. If
                provided that specific job ID is queried, otherwise
                a check for all outstanding jobs is checked.

        Returns:
            A list of dict objects with the status of the celery task/tasks
            that were outstanding.
        """
        if job_id:
            response = self.fetch_resource_data("tasks/?job_id={0:s}".format(job_id))
        else:
            response = self.fetch_resource_data("tasks/")

        return response.get("objects", [])

    def list_searchindices(self):
        """Yields all searchindices that the user has access to.

        Yields:
            A SearchIndex object instances.
        """
        response = self.fetch_resource_data("searchindices/")
        response_objects = response.get("objects")
        if not response_objects:
            yield None
            return

        for index_dict in response_objects[0]:
            index_id = index_dict["id"]
            index_name = index_dict["name"]
            index_obj = index.SearchIndex(
                searchindex_id=index_id, api=self, searchindex_name=index_name
            )
            yield index_obj

    def refresh_oauth_token(self):
        """Refresh an OAUTH token if one is defined."""
        if not self.credentials:
            return
        request = google.auth.transport.requests.Request()
        self.credentials.credential.refresh(request)

    def list_sigmarules(self, as_pandas=False):
        """Fetches Sigma rules from the database.
        Fetches all Sigma rules stored in the database on the system
        and returns a list of SigmaRule objects of the rules.

        Args:
            as_pandas: Boolean indicating that the results will be returned
                as a Pandas DataFrame instead of a list of SigmaRuleObjects.

        Returns:
            - List of Sigme rule object instances
            or
            - a pandas Dataframe with all rules if as_pandas is True.

        Raises:
            ValueError: If no rules are found.
        """
        rules = []
        response = self.fetch_resource_data("sigmarules/")

        if not response:
            raise ValueError("No rules found.")

        if as_pandas:
            return pandas.DataFrame.from_records(response.get("objects"))

        for rule_dict in response["objects"]:
            if not rule_dict:
                raise ValueError("No rules found.")

            index_obj = sigma.SigmaRule(api=self)
            for key, value in rule_dict.items():
                index_obj.set_value(key, value)
            rules.append(index_obj)
        return rules

    def create_sigmarule(self, rule_yaml):
        """Adds a single Sigma rule to the database.

        Adds a single Sigma rule to the database when `/sigmarules/` is called
        with a POST request.

        All attributes of the rule are taken by the `rule_yaml` value in the
        POST request.

        If no `rule_yaml` is found in the request, the method will fail as this
        is required to parse the rule.

        Args:
            rule_yaml: YAML of the Sigma Rule.

        Returns:
            Instance of a Sigma object.
        """

        retry_count = 0
        objects = None
        while True:
            resource_url = "{0:s}/sigmarules/".format(self.api_root)
            form_data = {"rule_yaml": rule_yaml}
            response = self.session.post(resource_url, json=form_data)
            response_dict = error.get_response_json(response, logger)
            objects = response_dict.get("objects")
            if objects:
                break
            retry_count += 1

            if retry_count >= self.DEFAULT_RETRY_COUNT:
                raise RuntimeError("Unable to create a new Sigma Rule.")

        rule_uuid = objects[0]["rule_uuid"]
        return self.get_sigmarule(rule_uuid)

    def get_sigmarule(self, rule_uuid):
        """Fetches a single Sigma rule from the database.
        Fetches a single Sigma rule selected by the `UUID`

        Args:
            rule_uuid: UUID of the Sigma rule.

        Returns:
            Instance of a SigmaRule object.
        """
        sigma_obj = sigma.SigmaRule(api=self)
        sigma_obj.from_rule_uuid(rule_uuid)

        return sigma_obj

    def parse_sigmarule_by_text(self, rule_text):
        """Obtain a parsed Sigma rule by providing text.

        Will parse a provided text `rule_yaml`, parse it and return as SigmaRule
        object.

        Args:
            rule_text: Full Sigma rule text.

        Returns:
            Instance of a Sigma object.

        Raises:
            ValueError: No Rule text given or issues parsing it.
        """
        if not rule_text:
            raise ValueError("No rule text given.")

        try:
            sigma_obj = sigma.Sigma(api=self)
            sigma_obj.from_text(rule_text)
        except ValueError:
            logger.error("Parsing Error, unable to parse the Sigma rule", exc_info=True)

        return sigma_obj  # pytype: disable=name-error  # py310-upgrade



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API sigma library."""
from __future__ import unicode_literals

import logging

from . import resource
from . import error

logger = logging.getLogger("timesketch_api.sigma")


class SigmaRule(resource.BaseResource):
    """Timesketch SigmaRule object.

    A SigmaRule object in Timesketch is a collection of one or more rules.

    Attributes:
        rule_uuid: The ID of the rule.
    """

    def __init__(self, api):
        """Initializes the Sigma object.

        Args:
            api: An instance of TimesketchApi object.

        """
        self._attr_dict = {}
        resource_uri = "sigmarules/"
        super().__init__(api=api, resource_uri=resource_uri)

    @property
    def attributes(self):
        """Returns a list of all attribute keys for the rule"""
        return list(self._attr_dict.keys())

    def get_attribute(self, key):
        """Get a value for a given key in case it has no dedicated property"""
        if not self._attr_dict:
            return ""
        return self._attr_dict.get(key, "")

    @property
    def search_query(self):
        """Returns the Search query."""
        return self.get_attribute("search_query")

    @property
    def title(self):
        """Returns the Sigma rule title."""
        return self.get_attribute("title")

    @property
    def id(self):
        """Returns the Sigma rule id."""
        return self.get_attribute("id")

    @property
    def rule_uuid(self):
        """Returns the rule id."""
        return self.get_attribute("id")

    @property
    def description(self):
        """Returns the rule description."""
        return self.get_attribute("description")

    @property
    def level(self):
        """Returns the rule confidence level."""
        return self.get_attribute("level")

    @property
    def falsepositives(self):
        """Returns the rule falsepositives."""
        return self.get_attribute("falsepositives")

    @property
    def author(self):
        """Returns the rule author."""
        return self.get_attribute("author")

    @property
    def date(self):
        """Returns the rule date."""
        return self.get_attribute("date")

    @property
    def modified(self):
        """Returns the rule modified date."""
        return self.get_attribute("modified")

    @property
    def logsource(self):
        """Returns the rule logsource."""
        return self.get_attribute("logsource")

    @property
    def detection(self):
        """Returns the rule detection."""
        return self.get_attribute("detection")

    @property
    def references(self):
        """Returns the rule references."""
        return self.get_attribute("references")

    @property
    def status(self):
        """Returns the rule status."""
        return self.get_attribute("status")

    def set_value(self, key, value):
        """Sets the value for a given key

        Args:
            key: key to set the value
            value: value to set

        """
        self._attr_dict[key] = value

    def _load_rule_dict(self, rule_dict):
        """Load a dict into a rule"""
        for key, value in rule_dict.items():
            self.set_value(key, value)

    def from_rule_uuid(self, rule_uuid):
        """Get a SigmaRule object from a rule UUID.

        Args:
            rule_uuid: Id of the sigma rule.

        """
        self.resource_uri = f"sigmarules/{rule_uuid}"

        self.lazyload_data(refresh_cache=True)
        objects = self.data.get("objects")
        if not objects:
            error_msg = "Unable to parse Sigma rule {0} with given text".format(
                rule_uuid
            )
            logger.error(error_msg)
            raise ValueError(error_msg)
        rule_dict = objects[0]
        for key, value in rule_dict.items():
            self.set_value(key, value)

    def from_text(self, rule_text):
        """Obtain a parsed Sigma rule by providing text.

        Args:
            rule_text: Rule text to be parsed.

        Raises:
            ValueError: If rule could not be parsed.
        """
        self.resource_uri = "{0:s}/sigmarules/text/".format(self.api.api_root)
        data = {"title": "Get_Sigma_by_text", "content": rule_text}
        response = self.api.session.post(self.resource_uri, json=data)
        response_dict = error.get_response_json(response, logger)

        objects = response_dict.get("objects")
        if not objects:
            error_msg = "Sigma rule Parsing error with provided rule"
            logger.warning(error_msg)
            raise ValueError(error_msg)

        rule_dict = objects[0]
        for key, value in rule_dict.items():
            self.set_value(key, value)

    def delete(self):
        """Deletes the Sigma rule from Timesketch."""
        if not self.get_attribute("id"):
            logger.warning(
                "Unable to delete the Sigma rule, it does not appear to be "
                "saved in the first place."
            )
            return False

        resource_url = f"{self.api.api_root}/sigmarules/{self.get_attribute('id')}"

        response = self.api.session.delete(resource_url)
        return error.check_return_status(response, logger)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API search object."""
import datetime
import json
import logging
import re

import pandas

from . import error
from . import resource
from . import searchtemplate


logger = logging.getLogger("timesketch_api.search")


class Chip:
    """Class definition for a query filter chip."""

    # The type of a chip that is defined.
    CHIP_TYPE = ""

    # The chip value defines what property or attribute of the
    # chip class will be used to generate the chip value.
    CHIP_VALUE = ""

    # The value of the chip field.
    CHIP_FIELD = ""

    def __init__(self):
        """Initialize the chip."""
        self._active = True
        self._operator = "must"
        self._chip_field = self.CHIP_FIELD

    @property
    def active(self):
        """A property that returns whether the chip is active or not."""
        return self._active

    @active.setter
    def active(self, active):
        """Decide whether the chip is active or disabled."""
        self._active = bool(active)

    @property
    def chip(self):
        """A property that returns the chip value."""
        return {
            "field": self._chip_field,
            "type": self.CHIP_TYPE,
            "operator": self._operator,
            "active": self._active,
            "value": getattr(self, self.CHIP_VALUE, ""),
        }

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        raise NotImplementedError

    def set_include(self):
        """Configure the chip so the content needs to be included in results."""
        self._operator = "must"

    def set_exclude(self):
        """Configure the chip so content needs to be excluded in results."""
        self._operator = "must_not"

    def set_optional(self):
        """Configure the chip so the content is optional in results."""
        self._operator = "should"

    def set_active(self):
        """Set the chip as active."""
        self._active = True

    def set_disable(self):
        """Disable the chip."""
        self._active = False


class DateIntervalChip(Chip):
    """A date interval chip."""

    CHIP_TYPE = "datetime_interval"
    CHIP_VALUE = "interval"

    _DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"
    _DATE_FORMAT_MICROSECONDS = "%Y-%m-%dT%H:%M:%S.%f"
    _DATE_ONLY_FORMAT = "%Y-%m-%d"

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._date = None
        self._before = 5
        self._after = 5
        self._unit = "m"

    def add_interval(self, before, after=None, unit="m"):
        """Set the interval of the chip.

        Args:
            before (int): the number of units that should be included
                before the date.
            after (int): optional number of units after the date. If not
                provided the value of before is used.
            unit (str): optional string with the unit of interval. This can
                be s for seconds, m for minutes, d for days and h for hours.
                The default value is m (minutes).

        Raises:
            ValueError if the unit is not correctly formed.
        """
        if after is None:
            after = before

        self.unit = unit

        self._before = before
        self._after = after

    @property
    def after(self):
        """Property that returns the time interval after the date."""
        return self._after

    @after.setter
    def after(self, after):
        """Make changes to the time interval after the date."""
        self._after = after

    @property
    def before(self):
        """Property that returns the time interval before the date."""
        return self._before

    @before.setter
    def before(self, before):
        """Make changes to the time interval before the date."""
        self._before = before

    @property
    def date(self):
        """Property that returns back the date."""
        if not self._date:
            return ""
        if self._date.microsecond == 0:
            return self._date.strftime(self._DATE_FORMAT)
        return self._date.strftime(self._DATE_FORMAT_MICROSECONDS)[:-3]

    @date.setter
    def date(self, date):
        """Make changes to the date."""
        try:
            dt = datetime.datetime.strptime(date, self._DATE_FORMAT_MICROSECONDS)
        except ValueError:
            try:
                dt = datetime.datetime.strptime(date, self._DATE_FORMAT)
            except ValueError:
                try:
                    dt = datetime.datetime.strptime(date, self._DATE_ONLY_FORMAT)
                except ValueError as exc:
                    logger.error(
                        "Unable to add date chip, wrong date format", exc_info=True
                    )
                    raise ValueError("Wrong date format") from exc
        self._date = dt

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        value = chip_dict.get("value")
        if not value:
            return

        split_value = value.split()
        if len(split_value) == 3:
            date_time, before, after = split_value
        elif len(split_value) == 4:
            date, time, before, after = split_value
            date_time = f"{date}T{time}"
        else:
            raise ValueError("Unable to configure date chip, wrong date format.")

        self.unit = before[-1]
        if date_time.endswith("Z"):
            self.date = date_time[:-1]
        else:
            self.date = date_time
        self.before = int(before[1:-1])
        self.after = int(after[1:-1])

    @property
    def interval(self):
        """A property that returns back the full interval."""
        return f"{self.date} -{self.before}{self.unit} +{self.after}{self.unit}"

    @property
    def unit(self):
        """Property that returns back the unit used."""
        return self._unit

    @unit.setter
    def unit(self, unit):
        """Make changes to the unit."""
        if unit not in ("s", "m", "d", "h"):
            raise ValueError(
                "Unable to add interval, needs to be one of: "
                "s (seconds), m (minutes), h (hours) or d (days)"
            )
        self._unit = unit


class DateRangeChip(Chip):
    """A date range chip."""

    CHIP_TYPE = "datetime_range"
    CHIP_VALUE = "date_range"

    _DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"
    _DATE_FORMAT_MICROSECONDS = "%Y-%m-%dT%H:%M:%S.%f"

    _DATE_RE = r"^[0-9]{4}-[0-9]{1,2}-[0-9]{2}$"

    def __init__(self):
        """Initialize the date range."""
        super().__init__()
        self._start_date = None
        self._end_date = None
        self._date_re = re.compile(self._DATE_RE)

    def add_end_time(self, end_time):
        """Add an end time to the range.

        Args:
            end_time (str): date string using the format '%Y-%m-%dT%H:%M:%s'

        Raises:
            ValueError: if the date format is incorrectly formatted.
        """
        if end_time.endswith("Z"):
            end_time = end_time[:-1]

        # Check for a whole day, YYYY-MM-DD.
        if self._date_re.match(end_time):
            end_time = f"{end_time}T23:59:59"

        try:
            dt = datetime.datetime.strptime(end_time, self._DATE_FORMAT_MICROSECONDS)
        except ValueError as exc:
            try:
                dt = datetime.datetime.strptime(end_time, self._DATE_FORMAT)
            except ValueError as exc:
                logger.error(
                    "Unable to add date chip, wrong date format", exc_info=True
                )
                raise ValueError("Wrong date format") from exc
        self._end_date = dt

    def add_start_time(self, start_time):
        """Add a start time to the range.

        Args:
            start_time (str): date string using the format '%Y-%m-%dT%H:%M:%s'

        Raises:
            ValueError: if the date format is incorrectly formatted.
        """
        if start_time.endswith("Z"):
            start_time = start_time[:-1]

        # Check for a whole day, YYYY-MM-DD.
        if self._date_re.match(start_time):
            start_time = f"{start_time}T00:00:00"

        try:
            dt = datetime.datetime.strptime(start_time, self._DATE_FORMAT_MICROSECONDS)
        except ValueError as exc:
            try:
                dt = datetime.datetime.strptime(start_time, self._DATE_FORMAT)
            except ValueError as exc:
                logger.error(
                    "Unable to add date chip, wrong date format", exc_info=True
                )
                raise ValueError("Wrong date format") from exc
        self._start_date = dt

    @property
    def end_time(self):
        """Property that returns the end time of a range."""
        if not self._end_date:
            return ""
        if self._end_date.microsecond == 0:
            return self._end_date.strftime(self._DATE_FORMAT)
        return self._end_date.strftime(self._DATE_FORMAT_MICROSECONDS)[:-3]

    @end_time.setter
    def end_time(self, end_time):
        """Sets the new end time."""
        self.add_end_time(end_time)

    @property
    def date_range(self):
        """Property that returns back the range."""
        return f"{self.start_time},{self.end_time}"

    @date_range.setter
    def date_range(self, date_range):
        """Sets the new range of the date range chip."""
        start_time, end_time = date_range.split(",")
        self.add_start_time(start_time)
        self.add_end_time(end_time)

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        chip_value = chip_dict.get("value")
        if not chip_value:
            return
        start, end = chip_value.split(",")
        self.start_time = start
        self.end_time = end

    @property
    def start_time(self):
        """Property that returns the start time of a range."""
        if not self._start_date:
            return ""
        if self._start_date.microsecond == 0:
            return self._start_date.strftime(self._DATE_FORMAT)
        return self._start_date.strftime(self._DATE_FORMAT_MICROSECONDS)[:-3]

    @start_time.setter
    def start_time(self, start_time):
        """Sets the new start time of a range."""
        self.add_start_time(start_time)


class LabelChip(Chip):
    """Label chip."""

    CHIP_TYPE = "label"
    CHIP_VALUE = "label"

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._label = ""

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        chip_value = chip_dict.get("value")
        if not chip_value:
            return

        self.label = chip_value

    @property
    def label(self):
        """Property that returns back the label."""
        return self._label

    @label.setter
    def label(self, label):
        """Make changes to the label."""
        self._label = label

    def use_comment_label(self):
        """Use the comment label."""
        self._label = "__ts_comment"

    def use_star_label(self):
        """Use the star label."""
        self._label = "__ts_star"


class TermChip(Chip):
    """Term chip definition."""

    CHIP_TYPE = "term"
    CHIP_VALUE = "query"

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._query = ""

    @property
    def field(self):
        """Property that returns back the field used to match against."""
        return self._chip_field

    @field.setter
    def field(self, field):
        """Make changes to the field used to match against."""
        self._chip_field = field

    def from_dict(self, chip_dict):
        """Configure the term chip from a dictionary."""
        chip_value = chip_dict.get("value")
        if not chip_value:
            return

        self.field = chip_dict.get("field")
        self.query = chip_value

    @property
    def query(self):
        """Property that returns back the query."""
        return self._query

    @query.setter
    def query(self, query):
        """Make changes to the query."""
        self._query = query


class Search(resource.SketchResource):
    """Search object."""

    DEFAULT_SIZE_LIMIT = 10000

    def __init__(self, sketch):
        resource_uri = f"sketches/{sketch.id}/explore/"
        super().__init__(sketch=sketch, resource_uri=resource_uri)

        self._aggregations = ""
        self._chips = []
        self._created_at = ""
        self._description = ""
        self._indices = "_all"
        self._max_entries = self.DEFAULT_SIZE_LIMIT
        self._name = ""
        self._query_dsl = ""
        self._query_filter = {}
        self._query_string = ""
        self._raw_response = None
        self._return_fields = ""
        self._scrolling = None
        self._searchtemplate = ""
        self._total_elastic_size = 0
        self._updated_at = ""

    def _extract_chips(self, query_filter):
        """Extract chips from a query_filter."""
        chips = query_filter.get("chips", [])
        if not chips:
            return

        for chip_dict in chips:
            chip_type = chip_dict.get("type")
            if not chip_type:
                continue

            chip = None
            if chip_type == "datetime_interval":
                chip = DateIntervalChip()
            elif chip_type == "datetime_range":
                chip = DateRangeChip()
            elif chip_type == "label":
                chip = LabelChip()
            elif chip_type == "term":
                chip = TermChip()

            if not chip:
                continue
            chip.from_dict(chip_dict)

            active = chip_dict.get("active", True)
            chip.active = active

            operator = chip_dict.get("operator", "must")
            if operator == "must":
                chip.set_include()
            elif operator == "must_not":
                chip.set_exclude()

            self.add_chip(chip)

    def _execute_query(self, file_name="", count=False):
        """Execute a search request and store the results.

        Args:
            file_name (str): Optional file path to a filename that
                all the results will be saved to. If not provided
                the results will be stored in the search object.
            count (bool): Optional boolean that determines whether
                we want to execute the query or only count the
                number of events that the query would produce. If
                set to True, the results will be stored in the
                search object, and the number of events will be
                returned.

        Returns:
            A dict with the search results or the total number of events
            (if count=True) or None if saved to file.
        """
        query_filter = self.query_filter
        if not isinstance(query_filter, dict):
            raise ValueError("Unable to query with a query filter that isn't a dict.")

        stop_size = self.max_entries
        scrolling = not bool(stop_size and (stop_size < self.DEFAULT_SIZE_LIMIT))

        if self.scrolling is not None:
            scrolling = self.scrolling

        form_data = {
            "query": self.query_string,
            "filter": query_filter,
            "dsl": self.query_dsl,
            "count": count,
            "fields": self.return_fields,
            "enable_scroll": scrolling,
            "file_name": file_name,
        }

        response = self.api.session.post(
            f"{self.api.api_root}/{self.resource_uri}", json=form_data
        )
        if not error.check_return_status(response, logger):
            error.error_message(
                response, message="Unable to query results", error=ValueError
            )

        if file_name:
            with open(file_name, "wb") as fw:
                fw.write(response.content)
            return None

        response_json = error.get_response_json(response, logger)

        if count:
            meta = response_json.get("meta", {})
            self._total_elastic_size = meta.get("total_count", 0)
            return meta.get("total_count", 0)

        scroll_id = response_json.get("meta", {}).get("scroll_id", "")
        form_data["scroll_id"] = scroll_id

        count = len(response_json.get("objects", []))
        total_count = count
        while count > 0:
            if self.max_entries and total_count >= self.max_entries:
                break

            if not scroll_id:
                logger.debug("No scroll ID, will stop.")
                break

            more_response = self.api.session.post(
                f"{self.api.api_root}/{self.resource_uri}", json=form_data
            )
            if not error.check_return_status(more_response, logger):
                error.error_message(
                    response, message="Unable to query results", error=ValueError
                )
            more_response_json = error.get_response_json(more_response, logger)
            count = len(more_response_json.get("objects", []))
            total_count += count
            response_json["objects"].extend(more_response_json.get("objects", []))
            more_meta = more_response_json.get("meta", {})
            added_time = more_meta.get("es_time", 0)
            response_json["meta"]["es_time"] += added_time

        self._total_elastic_size = response_json.get("meta", {}).get(
            "es_total_count", 0
        )
        if self._total_elastic_size != total_count:
            logger.info(
                "%d results were returned, but %d records matched the search query",
                total_count,
                self._total_elastic_size,
            )

        self._raw_response = response_json
        return response_json

    def add_chip(self, chip):
        """Add a chip to the ..."""
        self._chips.append(chip)
        self.commit()

    def add_date_range(self, start_time, end_time):
        """Add a date range chip to the search query.

        Args:
            start_time (str): a string with the start time of the range,
                the format should be '%Y-%m-%dT%H:%M:%S'
            end_time (str): a string with the end time of the range,
                the format should be '%Y-%m-%dT%H:%M:%S'
        """
        chip = DateRangeChip()
        chip.start_time = start_time
        chip.end_time = end_time
        self.add_chip(chip)

    @property
    def chips(self):
        """Property that returns all the chips in the search object."""
        return self._chips

    def commit(self):
        """Commit changes to the search object."""
        self._raw_response = None
        super().commit()

    @property
    def created_at(self):
        """Property that returns back the creation time of a search."""
        return self._created_at

    def delete(self):
        """Deletes the saved search from the store."""
        if not self._resource_id:
            logger.warning(
                "Unable to delete the saved search, it does not appear to be "
                "saved in the first place."
            )
            return False

        resource_url = (
            f"{self.api.api_root}/sketches/{self._sketch.id}/views/"
            f"{self._resource_id}/"
        )
        response = self.api.session.delete(resource_url)
        return error.check_return_status(response, logger)

    @property
    def description(self):
        """Property that returns back the description of the saved search."""
        return self._description

    @description.setter
    def description(self, description):
        """Make changes to the saved search description field."""
        self._description = description
        self.commit()

    @property
    def expected_size(self):
        """Property that returns the expected size of the search query."""
        if self._total_elastic_size:
            return self._total_elastic_size

        _ = self._execute_query(count=True)
        return self._total_elastic_size

    def from_manual(  # pylint: disable=arguments-differ
        self,
        query_string=None,
        query_dsl=None,
        query_filter=None,
        return_fields=None,
        max_entries=None,
        **kwargs,
    ):
        """Explore the sketch.

        Args:
            query_string (str): OpenSearch query string.
            query_dsl (str): OpenSearch query DSL as JSON string.
            query_filter (dict): Filter for the query as a dict.
            return_fields (str): A comma separated string with a list of fields
                that should be included in the response. Optional and defaults
                to None.
            max_entries (int): Optional integer denoting a best effort to limit
                the output size to the number of events. Events are read in,
                10k at a time so there may be more events in the answer back
                than this number denotes, this is a best effort.
            kwargs (dict[str, object]): Depending on the resource they may
                require different sets of arguments to be able to run a raw
                API request.

        Raises:
            ValueError: if unable to query for the results.
            RuntimeError: if the query is missing needed values, or if the
                sketch is archived.
        """
        super().from_manual(**kwargs)
        if not (query_string or query_filter or query_dsl):
            raise RuntimeError("You need to supply a query")

        self._username = self.api.current_user.username
        self._name = "From Explore"
        self._description = "From Explore"

        if query_filter:
            self.query_filter = query_filter

        self.query_string = query_string
        self.query_dsl = query_dsl
        self.return_fields = return_fields

        if max_entries:
            self.max_entries = max_entries

        # TODO: Make use of search templates and aggregations.
        # self._searchtemplate = data.get('searchtemplate', 0)
        # self._aggregations = data.get('aggregation', 0)

        self._created_at = datetime.datetime.now(datetime.timezone.utc).isoformat()
        self._updated_at = self._created_at

        self.resource_data = {}

    def from_saved(self, search_id):  # pylint: disable=arguments-differ
        """Initialize the search object from a saved search.

        Args:
            search_id: integer value for the saved
                search (primary key).
        """
        resource_uri = f"sketches/{self._sketch.id}/views/{search_id}/"
        resource_data = self.api.fetch_resource_data(resource_uri)

        data = resource_data.get("objects", [None])[0]
        if not data:
            logger.error("Unable to get any data back from a saved search.")
            return

        label_string = data.get("label_string", "")
        if label_string:
            self._labels = json.loads(label_string)
        else:
            self._labels = []

        self._aggregations = data.get("aggregation", 0)
        self._created_at = data.get("created_at", "")
        self._description = data.get("description", "")
        self._name = data.get("name", "")
        self.query_dsl = data.get("query_dsl", "")
        query_filter = data.get("query_filter", "")
        if query_filter:
            filter_dict = json.loads(query_filter)
            if "fields" in filter_dict:
                fields = filter_dict.pop("fields")
                return_fields = [x.get("field") for x in fields]
                self.return_fields = ",".join(return_fields)

            indices = filter_dict.get("indices", [])
            if indices:
                self.indices = indices

            self.query_filter = filter_dict
        self.query_string = data.get("query_string", "")
        self._resource_id = search_id
        self._searchtemplate = data.get("searchtemplate", 0)
        self._updated_at = data.get("updated_at", "")
        self._username = data.get("user", {}).get("username", "System")

        self.resource_data = data

    @property
    def indices(self):
        """Return the current set of indices used in the search."""
        return self._indices

    @indices.setter
    def indices(self, indices):
        """Make changes to the current set of indices."""
        if indices == "_all":
            self._indices = "_all"
            self.commit()
            return

        def _is_string_or_int(item):
            return isinstance(item, (str, int))

        if not isinstance(indices, list):
            logger.warning(
                "Indices needs to be a list of strings (indices that were "
                "passed in were not a list)."
            )
            return
        if not all(map(_is_string_or_int, indices)):
            logger.warning(
                "Indices needs to be a list of strings or ints, not all "
                "entries in the indices list are valid string/int."
            )
            return

        if len(indices) == 1 and indices[0] == "_all":
            self._indices = "_all"
            self.commit()
            return

        # Indices here can be either a list of timeline names, IDs or a list
        # of search indices. We need to verify that these exist before saving
        # them.
        valid_ids = set()
        timeline_indices = {}  # Dict[index] = List[name]
        timeline_names = {}  # Dict[name] = id

        for timeline_object in self._sketch.list_timelines():
            timeline_indices.setdefault(timeline_object.index_name, [])
            timeline_indices[timeline_object.index_name].append(timeline_object.name)
            valid_ids.add(timeline_object.id)

            timeline_names[timeline_object.name] = timeline_object.id

        new_indices = []
        for index in indices:
            # Is this an index name, include all timeline IDs.
            if index in timeline_indices:
                new_indices.extend(timeline_indices[index])
                continue

            # Is this a timeline ID?
            if isinstance(index, int):
                if index in valid_ids:
                    new_indices.append(str(index))
                    continue

            if isinstance(index, str):
                if index.isdigit():
                    if int(index) in valid_ids:
                        new_indices.append(index)
                        continue

            # Is this a timeline name?
            if index in timeline_names:
                new_indices.append(timeline_names[index])

        if not new_indices:
            logger.warning("No valid indices found, not changing the value.")
            return

        self._indices = new_indices
        self.commit()

    @property
    def max_entries(self):
        """Return the maximum number of entries in the return value."""
        return self._max_entries

    @max_entries.setter
    def max_entries(self, max_entries):
        """Make changes to the max entries of return values."""
        self._max_entries = max_entries
        if max_entries < self.DEFAULT_SIZE_LIMIT:
            _ = self.query_filter
            self._query_filter["size"] = max_entries
            self._query_filter["terminate_after"] = max_entries
        self.commit()

    @property
    def name(self):
        """Property that returns the query name."""
        return self._name

    @name.setter
    def name(self, name):
        """Make changes to the saved search name."""
        self._name = name
        self.commit()

    def order_ascending(self):
        """Set the order of objects returned back ascending."""
        # Trigger a creation of a query filter if it does not exist.
        _ = self.query_filter
        self._query_filter["order"] = "asc"
        self.commit()

    def order_descending(self):
        """Set the order of objects returned back descending."""
        # Trigger a creation of a query filter if it does not exist.
        _ = self.query_filter
        self._query_filter["order"] = "desc"
        self.commit()

    @property
    def query_dsl(self):
        """Property that returns back the query DSL."""
        return self._query_dsl

    @query_dsl.setter
    def query_dsl(self, query_dsl):
        """Make changes to the query DSL of the search."""
        if query_dsl and isinstance(query_dsl, str):
            query_dsl = json.loads(query_dsl)

        # Special condition of an empty DSL.
        if query_dsl == '""':
            query_dsl = ""

        self._query_dsl = query_dsl
        self.commit()

    @property
    def query_filter(self):
        """Property that returns the query filter."""
        if not self._query_filter:
            self._query_filter = {
                "size": self.DEFAULT_SIZE_LIMIT,
                "terminate_after": self.DEFAULT_SIZE_LIMIT,
                "indices": self.indices,
                "order": "asc",
                "chips": [],
            }

        query_filter = self._query_filter
        query_filter["chips"] = [x.chip for x in self._chips]
        query_filter["indices"] = self.indices
        return query_filter

    @query_filter.setter
    def query_filter(self, query_filter):
        """Make changes to the query filter."""
        if isinstance(query_filter, str):
            try:
                query_filter = json.loads(query_filter)
            except json.JSONDecodeError as exc:
                raise ValueError("Unable to parse the string as JSON") from exc

        if not isinstance(query_filter, dict):
            raise ValueError("Query filter needs to be a dict.")
        self._query_filter = query_filter
        self._extract_chips(query_filter)
        self.commit()

    @property
    def query_string(self):
        """Property that returns back the query string."""
        return self._query_string

    @query_string.setter
    def query_string(self, query_string):
        """Make changes to the query string of a saved search."""
        self._query_string = query_string
        self.commit()

    def remove_chip(self, chip_index):
        """Remove a chip from the saved search."""
        chip_len = len(self._chips)
        if chip_index > (chip_len + 1):
            raise ValueError(
                f"Unable to remove chip, only {chip_len} chips stored "
                f"(no index {chip_index})"
            )

        try:
            _ = self._chips.pop(chip_index)
        except IndexError as exc:
            raise ValueError(
                f"Unable to remove index {chip_index}, out of range"
            ) from exc

        self.commit()

    @property
    def return_fields(self):
        """Property that returns the return_fields."""
        if self._return_fields:
            items = self._return_fields.split(",")
            if "datetime" not in items:
                items.append("datetime")
            return ",".join(items)
        return self._return_fields

    @return_fields.setter
    def return_fields(self, return_fields):
        """Make changes to the return fields."""
        self._return_fields = return_fields
        self.commit()

    @property
    def return_size(self):
        """Return the maximum number of entries in the return value."""
        return self._max_entries

    @return_size.setter
    def return_size(self, return_size):
        """Make changes to the maximum number of entries in the return."""
        self._max_entries = return_size
        self.commit()

    def save(self):
        """Save the search in the database.

        Returns:
            String with the identification of the saved search.

        Raises:
            ValueError: if there are values missing in order to save the query.
            RuntimeError: if the search could not be saved.
        """
        if not self.name:
            raise ValueError("No name for the query saved. Please select a name first.")

        if not (self.query_string or self.query_dsl):
            raise ValueError(
                "Need to have either a query DSL or a query string to be "
                "able to save the search."
            )

        if not self.description:
            logger.warning("No description selected for search, saving without one")

        if self._resource_id:
            resource_url = (
                f"{self.api.api_root}/sketches/{self._sketch.id}/views/"
                f"{self._resource_id}/"
            )
        else:
            resource_url = f"{self.api.api_root}/sketches/{self._sketch.id}/views/"

        query_filter = self.query_filter
        if self.return_fields:
            sketch_data = self._sketch.data
            sketch_meta = sketch_data.get("meta", {})
            mappings = sketch_meta.get("mappings", [])

            use_mappings = []
            for field in self.return_fields.split(","):
                field = field.strip().lower()
                for map_entry in mappings:
                    if map_entry.get("field", "").lower() == field:
                        use_mappings.append(map_entry)
            query_filter["fields"] = use_mappings

        data = {
            "name": self.name,
            "description": self.description,
            "query": self.query_string,
            "filter": query_filter,
            "dsl": self.query_dsl,
            "labels": json.dumps(self.labels),
        }
        response = self.api.session.post(resource_url, json=data)
        status = error.check_return_status(response, logger)
        if not status:
            error.error_message(response, "Unable to save search", error=RuntimeError)

        response_json = error.get_response_json(response, logger)
        search_dict = response_json.get("objects", [{}])[0]
        self._resource_id = search_dict.get("id", 0)
        return f"Saved search to ID: {self._resource_id}"

    def save_as_template(self):
        """Save the search as a search template.

        Returns:
            A search template object (searchtemplate.SearchTemplate).
        """
        if not self._resource_id:
            logger.warning("Search has not been saved first, saving now.")
            return_string = self.save()
            logger.info(return_string)

        template = searchtemplate.SearchTemplate(self.api)
        template.from_search_object(self)

        print(template.save())
        self._searchtemplate = template.id
        return template

    @property
    def scrolling(self):
        """Returns whether scrolling is enabled or not."""
        return self._scrolling

    def scrolling_disable(self):
        """ "Disables scrolling."""
        self._scrolling = False

    def scrolling_enable(self):
        """Enable scrolling."""
        self._scrolling = True

    def to_dict(self):
        """Returns a dict with the respone of the query."""
        if self._raw_response is None:
            self._execute_query()
            if self._raw_response is None:
                raise ValueError("No results to return.")

        return self._raw_response

    def to_file(self, file_name):
        """Saves the content of the query to a file.

        Args:
            file_name (str): Full path to a file that will store the results
                of the query to as a ZIP file. The ZIP file will contain a
                METADATA file and a CSV with the results from the query.

        Returns:
            Boolean that determines if it was successful.
        """
        old_scrolling = self.scrolling
        self._scrolling = True
        self._execute_query(file_name=file_name)
        self._scrolling = old_scrolling
        return True

    def to_pandas(self):
        """Returns a pandas DataFrame with the response of the query."""
        if self._raw_response is None:
            self._raw_response = self._execute_query()
            if self._raw_response is None:
                raise ValueError("No results to return.")

        return_list = []
        timelines = {t.id: t.name for t in self._sketch.list_timelines()}

        return_field_list = []
        return_fields = self.return_fields
        if return_fields:
            if return_fields.startswith("'"):
                return_fields = return_fields[1:]
            if return_fields.endswith("'"):
                return_fields = return_fields[:-1]
            return_field_list = return_fields.split(",")

        for result in self._raw_response.get("objects", []):
            source = result.get("_source", {})
            if not return_fields or "_id" in return_field_list:
                source["_id"] = result.get("_id")
            if not return_fields or "_type" in return_field_list:
                source["_type"] = result.get("_type")
            if not return_fields or "_index" in return_field_list:
                source["_index"] = result.get("_index")
            if not return_fields or "_source" in return_field_list:
                source["_source"] = timelines.get(result.get("__ts_timeline_id"))
            if not return_fields or "__ts_timeline_id" in return_field_list:
                source["_source"] = timelines.get(result.get("__ts_timeline_id"))

            return_list.append(source)

        data_frame = pandas.DataFrame(return_list)
        if "datetime" in data_frame:
            try:
                data_frame["datetime"] = pandas.to_datetime(data_frame.datetime)
            except pandas.errors.OutOfBoundsDatetime:
                pass
        elif "timestamp" in data_frame:
            try:
                data_frame["datetime"] = pandas.to_datetime(
                    data_frame.timestamp / 1e6, utc=True, unit="s"
                )
            except pandas.errors.OutOfBoundsDatetime:
                pass

        return data_frame

    @property
    def updated_at(self):
        """Property that returns back the updated time of a search."""
        return self._updated_at



# Copyright 2023 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Timesketch API aggregation object."""
import unittest
import mock

import altair as alt

from . import aggregation
from . import client
from . import test_lib


class AggregationTest(unittest.TestCase):
    """Test Graph object."""

    @mock.patch("requests.Session", test_lib.mock_session)
    def setUp(self):
        """Setup test case."""
        self.api_client = client.TimesketchApi("http://127.0.0.1", "test", "test")
        self.sketch = self.api_client.get_sketch(1)

    # pylint: disable=protected-access
    def test_init(self):
        """Tests the Aggregation init method."""
        aggregation_obj = aggregation.Aggregation(self.sketch)

        # Aggregation attributes
        self.assertEqual(aggregation_obj._created_at, "")
        self.assertEqual(aggregation_obj._group_id, None)
        self.assertEqual(aggregation_obj._parameters, {})
        self.assertEqual(aggregation_obj._updated_at, "")
        self.assertEqual(aggregation_obj._aggregator_name, "")
        self.assertEqual(aggregation_obj.chart_color, "")
        self.assertEqual(aggregation_obj.chart_title, "")
        self.assertEqual(aggregation_obj.search_id, None)
        self.assertEqual(aggregation_obj.type, None)

        # SketchResource attributes
        self.assertEqual(aggregation_obj._labels, [])
        self.assertEqual(aggregation_obj._resource_id, 0)
        self.assertEqual(aggregation_obj._sketch, self.sketch)
        self.assertEqual(aggregation_obj._username, "")

        # BaseResource attributes
        self.assertEqual(aggregation_obj.api, self.api_client)
        self.assertEqual(
            aggregation_obj.resource_uri, "sketches/1/aggregation/explore/"
        )
        self.assertEqual(aggregation_obj.resource_data, None)

    def test_created_at(self):
        """Tests the created_at property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.created_at, "")

        aggregation_obj.from_saved(1)
        self.assertEqual(aggregation_obj.created_at, "2023-01-08T08:45:23.113454")

    def test_parameters(self):
        """Tests the parameters property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.parameters, {})

        aggregation_obj.from_saved(1)
        self.assertEqual(
            aggregation_obj.parameters,
            {
                "supported_charts": "barchart",
                "field": "ip",
                "start_time": "",
                "end_time": "",
                "limit": "10",
                "index": [1, 2],
            },
        )

    def test_is_part_of_group(self):
        """Tests the is_part_of_group property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.is_part_of_group, False)

        aggregation_obj.from_saved(1)
        self.assertEqual(aggregation_obj.is_part_of_group, False)

    def test_title(self):
        """Tests the title property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.title, "")

        aggregation_obj.from_saved(1)
        self.assertEqual(aggregation_obj.title, "Top results for an unknown field")

    def test_title_setter(self):
        """Tests the title setter."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        aggregation_obj.title = "new title"
        self.assertEqual(aggregation_obj.title, "new title")

    def test_chart(self):
        """Tests the chart property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        with self.assertRaises(TypeError) as err:
            _ = aggregation_obj.chart
        self.assertEqual(
            str(err.exception), "Unable to generate chart, missing a chart type."
        )

        aggregation_obj.from_saved(1)
        chart = aggregation_obj.chart
        self.assertIsInstance(chart, alt.Chart)

    def test_description(self):
        """Tests the description property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.description, "")

        aggregation_obj.from_saved(1)
        self.assertEqual(
            aggregation_obj.description, "Aggregating values of a particular field"
        )

    def test_description_setter(self):
        """Tests the description property setter."""

    def test_name(self):
        """Tests the name property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.name, "")

        aggregation_obj.from_saved(1)
        self.assertEqual(aggregation_obj.name, "ip barchart")

    def test_aggregator_name(self):
        """Tests the aggregator_name property."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        self.assertEqual(aggregation_obj.name, "")

        aggregation_obj.from_saved(1)
        self.assertEqual(aggregation_obj.name, "ip barchart")

    def test_add_label(self):
        """Test the add_label function."""

    def test_from_saved(self):
        """Tests the from_saved method."""
        aggregation_obj = aggregation.Aggregation(self.sketch)
        aggregation_obj.from_saved(2)
        self.assertEqual(aggregation_obj._aggregator_name, "field_bucket")
        self.assertEqual(aggregation_obj.chart_color, "")
        self.assertEqual(aggregation_obj.chart_type, "table")
        self.assertEqual(aggregation_obj.chart_title, "")
        self.assertEqual(aggregation_obj.search_id, None)
        self.assertEqual(aggregation_obj.type, "stored")

        self.assertEqual(aggregation_obj.created_at, "2023-01-08T08:46:24.871292")
        self.assertEqual(
            aggregation_obj.parameters,
            {
                "supported_charts": "table",
                "field": "domain",
                "start_time": "",
                "end_time": "",
                "limit": "10",
                "index": [1, 2],
            },
        )
        self.assertEqual(aggregation_obj.is_part_of_group, False)
        self.assertEqual(aggregation_obj.title, "Top results for an unknown field")
        self.assertEqual(
            aggregation_obj.description, "Aggregating values of a particular field"
        )
        self.assertEqual(aggregation_obj.name, "domain table")



#!/usr/bin/env python
# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This is the setup file for the project."""
from __future__ import unicode_literals

from setuptools import find_packages
from setuptools import setup

from timesketch_api_client import version


long_description = (
    "The Timesketch API client provides you with a set of Python libraries "
    "to connect to your Timesketch (https://github.com/google/timesketch) "
    "instance.\n\n"
    "The API is feature complete with the Timesketch UI and allows you to "
    "do all operations that can be done in the UI, providing ways to "
    "integrate Timesketch into other products such as Jupyter/Colab.\n\n"
    "To see how it works in action, try the colab notebook that is accessible "
    "from here: https://colab.research.google.com/github/google/timesketch/"
    "blob/master/notebooks/colab-timesketch-demo.ipynb"
)


setup(
    name="timesketch-api-client",
    version=version.get_version(),
    description="Timesketch API client",
    long_description=long_description,
    license="Apache License, Version 2.0",
    url="http://www.timesketch.org/",
    maintainer="Timesketch development team",
    maintainer_email="timesketch-dev@googlegroups.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=frozenset(
        [
            "pandas",
            "cryptography",
            "requests",
            "altair",
            "google-auth",
            "networkx",
            "google_auth_oauthlib",
            "beautifulsoup4",
        ]
    ),
)
